//------------------------------------------------
// Collision.c 蔻菈
//------------------------------------------------

#include "funcs.h"
#include <math.h>
#include <stdlib.h>

#ifndef __Collision__
#define __Collision__

// `
#define EPS      1e-10		// ̌덷el

#define sinf sin
#define cosf cos



//-----------------------------------------------------------------
// ~̍\
// xNgWƔaȂ
// pr
//-----------------------------------------------------------------
typedef struct _Circle
{
	Vector xy;	// SxNg
	int    r;	// a
} Circle;





//-----------------------------------------------------------------
// LineFromVector
// @\F2xNg𐶐
// FVector *v1, Vector *v2
//       *v1 : n_ʒuxNg
//       *v2 : I_ʒuxNg
// ԒlFꂽ
Line LineFromVector(Vector *v1, Vector *v2)
{
	Line ret;
	ret.st = *v1;
	ret.ed = *v2;
	return ret;
}

//-----------------------------------------------------------------
// LineFromXY
// @\FQ_𐶐
// Ffloat x1, float y1, float x2, float y2
//       x1, y1 : n_W
//       x2, y2 : I_W
// ԒlFꂽ
Line LineFromXY(float x1, float y1, float x2, float y2)
{
	Line ret;
	ret.st.x = x1-CenterX;
	ret.st.y = y1-CenterY;
	ret.ed.x = x2-CenterX;
	ret.ed.y = y2-CenterY;
	return ret;
}

//-----------------------------------------------------------------
// LineRotate
// @\F]
// FLine *line, float rad
//       *line : ]AhX
//       rad   : ]pxiWAj
// ԒlFȂ
void LineRotate(Line *line, float rad)
{
	float sinx = sinf(rad),
	      cosx = cosf(rad);
	float x;
	         x = line->st.x*cosx - line->st.y*sinx,
	line->st.y = line->st.x*sinx + line->st.y*cosx;
	line->st.x = x;
	         x = line->ed.x*cosx - line->ed.y*sinx,
	line->ed.y = line->ed.x*sinx + line->ed.y*cosx;
	line->ed.x = x;
}

//-----------------------------------------------------------------
// LineRotateArray
// @\Fz]
// FLine *line, int num, float rad
//       *line : ]zAhX
//       num   : 
//       rad   : ]pxiWAj
// ԒlFȂ
void LineRotateArray(Line *line, int num, float rad)
{
	float sinx = sinf(rad),
	      cosx = cosf(rad);
	float x;
	int i;
	if(rad == 0)	return;
	
	for(i=0; i<num; i++){
		           x = line[i].st.x*cosx - line[i].st.y*sinx,
		line[i].st.y = line[i].st.x*sinx + line[i].st.y*cosx;
		line[i].st.x = x;
		           x = line[i].ed.x*cosx - line[i].ed.y*sinx,
		line[i].ed.y = line[i].ed.x*sinx + line[i].ed.y*cosx;
		line[i].ed.x = x;
	}
}

//-----------------------------------------------------------------
// LineCrossCheck
// @\F2̌𔻒肷
// FLine *line1, Line *line2
//       *line1 : P̃AhX
//       *line2 : Q̃AhX
// ԒlFtrue  : 
//       false : Ȃis܂ށj
// QlFhttp://www5d.biglobe.ne.jp/~tomoya03/shtml/algorithm/IntersectionEX.htm
bool LineCrossCheck(Line *line1, Line *line2)
{
	// 
	if( (line1->st.x - line1->ed.x) * (line2->st.y - line1->st.y) +
	    (line1->st.y - line1->ed.y) * (line1->st.x - line2->st.x) *
	    (line1->st.x - line1->ed.x) * (line2->ed.y - line1->st.y) +
	    (line1->st.y - line1->ed.y) * (line1->st.x - line2->ed.x) > 0 )
		return false;
	if( (line2->st.x - line2->ed.x) * (line1->st.y - line2->st.y) +
	    (line2->st.y - line2->ed.y) * (line2->st.x - line1->st.x) *
	    (line2->st.x - line2->ed.x) * (line1->ed.y - line2->st.y) +
	    (line2->st.y - line2->ed.y) * (line2->st.x - line1->ed.x) > 0 )
		return false;

	// ̑SĂ̔NAĂ
	return true;
}

//-----------------------------------------------------------------
// LineCrossGet
// @\F2̌_߂
// FLine *line1, Line *line2, Vector *vector
//       *line1  : P̃AhX
//       *line2  : Q̃AhX
//       *vector : ʂi[xNg̃AhX
// ԒlFȂ
// QlFhttp://www.deqnotes.net/acmicpc/2d_geometry/lines
void LineCrossGet(Line *line1, Line *line2, Vector *vector)
{
	Vector b  = VectorSubNew(&line2->ed, &line2->st),
	       v1 = VectorSubNew(&line1->st, &line2->st),
	       v2 = VectorSubNew(&line1->ed, &line2->st);
	float  f1 = fabs(VectorCrossProduct(&b, &v1)),
	       f2 = fabs(VectorCrossProduct(&b, &v2)),
	       t  = f1 / (f1 + f2);
	*vector = VectorSubNew(&line1->ed, &line1->st);
	VectorMult(vector, t);
	VectorAdd(vector, &line1->st);
}


//-----------------------------------------------------------------
// LineDotLength
// @\FƓ_Ƃ̍ŏ߂
// FLine *line, Vector *vector, int *type
//       *line    : 
//       *vector  : _̃|C^
//       *type    : ŒZ̏
//                  type = 0 : ɍŒZ
//                         1 : n_ŒZ
//                         2 : I_ŒZ
// ԒlFŒZ̒l
// QlFhttp://www.deqnotes.net/acmicpc/2d_geometry/lines
float LineDotLength(Line *line, Vector *vector, int *type)
{
	Vector v1 = VectorSubNew(&line->ed, &line->st),	// b-a
	       v2 = VectorSubNew(vector, &line->st),	// c-a
	       v3 = VectorInvertNew(&v1),		// a-b
	       v4 = VectorSubNew(vector, &line->ed);	// c-b
	if(VectorDotProduct(&v1, &v2) < EPS){
		*type = 1;
		return VectorLength(&v2);
	}
	if(VectorDotProduct(&v3, &v4) < EPS){
		*type = 2;
		return VectorLength(&v4);
	}
	*type = 0;
	return fabs(VectorCrossProduct(&v1, &v2))/VectorLength(&v1);
}

//-----------------------------------------------------------------
// LineCircleTime
// @\Fړ~̊OڂƐ̐ڐG
// FMyObject *obj, Line *line, vector *vy
//       *obj  : ~̏
//       *line : 
//       *vy   : ړxNg
// ԒlFڐGԁiPt[PƂĊōlj
//       ڐGȂꍇ͕Ԃ
// QlFhttp://marupeke296.com/COL_3D_No16_MoveShereAndSepLine.html
float LineCircleTime(MyObject *obj, Line *line, Vector *vy)
{
	float a, t, b, o;
	Vector A, C;
	
	if(vy->x==0 && vy->y==0) return -1;
	
	// Ql̂`߂
	A = line->ed;
	a = VectorDotProduct(vy, &A);
	VectorMult(&A, a);
	VectorSub(&A, vy);
	
	// Ql̂b߂
	C = VectorSubNew(&obj->xy, &line->st);
	VectorMult(&C, VectorDotProduct(&C, &line->ed));
	VectorAdd(&C, &line->st);
	VectorSub(&C, &obj->xy);
	
	// 
	a = VectorLengthNoSquare(&A);
	b = VectorDotProduct(&A, &C);
	o = VectorLengthNoSquare(&C) - obj->r*obj->r;
	t = b*b-a*o;
	
	if(t < 0)	return -1;
	if(t == 0)	return -b/a;
	return (-b-sqrt(t))/a;
}

//-----------------------------------------------------------------
// @\FNCbN\[gpr֐
//       ̍ɗp
// FVector *v1, Vector *v2
// ԒlFr
// QlFhttp://www.bohyoh.com/CandCPP/FAQ/FAQ00047.html
// ⑫Fn_yWɂď
int LineCmpYst(const Line *line1, const Line *line2)
{
    return (line1->st.y < line2->st.y ? -1 :
            line1->st.y > line2->st.y ?  1 : 0);
}
// ⑫FI_yWɂď
int LineCmpYed(const Line *line1, const Line *line2)
{
    return (line1->ed.y < line2->ed.y ? -1 :
            line1->ed.y > line2->ed.y ?  1 : 0);
}
// ⑫Fn_xWɂď
int LineCmpXst(const Line *line1, const Line *line2)
{
    return (line1->st.x < line2->st.x ? -1 :
            line1->st.x > line2->st.x ?  1 : 0);
}
// ⑫FI_xWɂď
int LineCmpXed(const Line *line1, const Line *line2)
{
    return (line1->ed.x < line2->ed.x ? -1 :
            line1->ed.x > line2->ed.x ?  1 : 0);
}

//-----------------------------------------------------------------
// LineUpSwap
// @\F蔻Cn_yWɗ悤ɕёւ
// FLine *line, int num;
//       *line : z
//       num   : z̗vf
// ԒlFȂ
void LineUpSwap(Line *line, int num)
{
	int i;
	Vector tmp;
	for(i=0; i<num; i++){
		if(line[i].st.y > line[i].ed.y){
			tmp          = line[i].st;
			line[i].st = line[i].ed;
			line[i].ed = tmp;
		}
	}
}

//-----------------------------------------------------------------
// LineLeftSwap
// @\F蔻Cn_xWɗ悤ɕёւ
// FLine *line, int num;
//       *line : z
//       num   : z̗vf
// ԒlFȂ
void LineLeftSwap(Line *line, int num)
{
	int i;
	Vector tmp;
	for(i=0; i<num; i++){
		if(line[i].st.x > line[i].ed.x){
			tmp        = line[i].st;
			line[i].st = line[i].ed;
			line[i].ed = tmp;
		}
	}
}

//-----------------------------------------------------------------
// LineLowerBound
// @\Fw肳ꂽlȏ̒lŏɂłvfԍԂ
//       񕪒TpD
// ӁFz͏\[gĂKv
// FLine *line, int array, float num
//       *line : Cz
//       array : z̗vf
//       num   : rɗpl
// ԒlFl傫ŏ̗vfԍ
//       Ȃꍇ-1Ԃ
// ⑫FI_xWɂĒT
int LineLowerBoundXed(Line *line, int array, float num)
{
	int start = 0,
	    end   = array-1,
	    center;
	
	if(num < line[start].ed.x) return 0;	// ŏlTl
	if(num > line[end].ed.x) return -1;	// őlTl傫
	
	while(start != end){
		center = (start+end)>>1;
		if(line[center].ed.x < num)
			start = center;
		else
			end = center;
	}
	return start;
}
// ⑫FI_yWɂĒT
int LineLowerBoundYed(Line *line, int array, float num)
{
	int start = 0,
	    end   = array-1,
	    center;
	
	if(num < line[start].ed.y) return 0;	// ŏlTl
	if(num > line[end].ed.y) return -1;	// őlTl傫
	
	while(end-start > 1){
		center = (start+end)>>1;
		if(line[center].ed.y < num){
			start = center;
		} else {
			end = center;
		}
	}
	if(end-start == 1)
		return start>num ? start : end;
	return start;
}

//-----------------------------------------------------------------
// LineUpperBound
// @\Fw肳ꂽlȉ̒lŏɂłvfԍԂ
//       񕪒TpD
// ӁFz͏\[gĂKv
// FLine *line, int array, float num
//       *line : Cz
//       array : z̗vf
//       num   : rɗpl
// ԒlFl菬ő̗vfԍ
//       Ȃꍇ-1Ԃ
// ⑫Fn_xWɂĒT
int LineUpperBoundXst(Line *line, int array, float num)
{
	int start = 0,
	    end   = array-1,
	    center;
	
	if(num < line[start].st.x) return -1;	// ŏlTl
	if(num > line[end].st.x) return end;	// őlTl傫
	
	while(end-start > 1){
		center = (start+end)>>1;
		if(line[center].st.x < num){
			start = center;
		} else {
			end = center;
		}
	}
	if(end-start == 1)
		return start>num ? start : end;
	return start;
}
// ⑫Fn_yWɂĒT
int LineUpperBoundYst(Line *line, int array, float num)
{
	int start = 0,
	    end   = array-1,
	    center;
	
	if(num < line[start].st.y) return -1;	// ŏlTl
	if(num > line[end].st.y) return end;	// őlTl傫
	
	while(end-start > 1){
		center = (start+end)>>1;
		if(line[center].st.y < num){
			start = center;
		} else {
			end = center;
		}
	}
	if(end-start == 1)
		return start>num ? start : end;
	return start;
}

//-----------------------------------------------------------------
// CollisionAreaDelete
// @\FsKvȗ̈폜
// lFCollisionProcess̕ɂč쐬
//       line ɂ͒Tׂŏ̗vf
//       num ɂ͒Tׂvf
// ԒlFȂ
void CollisionAreaDelete(MyObject *obj, Line **line, int *num){
	
	int   start = 0,	// ̈̎n
	      end = *num-1,	// ̈̏I
	      tmp;
	float upm,		// ̈폜p̏̒l
	      downm,		//             ̒l
	      leftm,		//             ̒l
	      rightm;		//             E̒l
	
	if(obj->velocity.x > 0){
		leftm  = obj->xy.x - obj->r;
		rightm = obj->xy.x + obj->r + obj->velocity.x;
	} else {
		leftm  = obj->xy.x - obj->r + obj->velocity.x;
		rightm = obj->xy.x + obj->r;
	}
	if(obj->velocity.y > 0){
		upm    = obj->xy.y - obj->r;
		downm  = obj->xy.y + obj->r + obj->velocity.y;
	} else {
		upm    = obj->xy.y - obj->r + obj->velocity.y;
		downm  = obj->xy.y + obj->r;
	}
		

	// ̈폜
	LineUpSwap(*line, *num);
	qsort(*line, *num, sizeof(Line), (int(*)(const void*, const void*))LineCmpYed);
	tmp    = LineLowerBoundYed(*line, *num, upm);
	if(tmp != -1){
		*num  -= tmp;
		*line += tmp;
		start = tmp;
	}
	
	// ̈폜
	qsort(*line, *num, sizeof(Line), (int(*)(const void*, const void*))LineCmpYst);
	tmp    = LineUpperBoundYst(*line, *num, downm);
	if(tmp != -1){
		*num -= *num-tmp;
		end   = tmp;
	}
	
	// ̈폜 
	LineUpSwap(*line, *num);
	qsort(*line, *num, sizeof(Line), (int(*)(const void*, const void*))LineCmpXed);
	tmp  = LineLowerBoundXed(*line, *num, leftm);
	if(tmp != -1){
		*num  -= tmp;
		*line += tmp;
		start  = tmp;
	}
	
	// Ë폜
	qsort(*line, *num, sizeof(Line), (int(*)(const void*, const void*))LineCmpXst);
	tmp   = LineUpperBoundXst(*line, *num, rightm);
	if(tmp != -1){
		*num -= *num-tmp;
		end   = tmp;
	}
}

//-----------------------------------------------------------------
// CollisionProcess
// @\F@̓蔻
// FMyObject *obj, Line *line, int num
//       *obj  : @
//       *line : Cz
//       num   : z̗vf
// ԒlFȂ
void CollisionProcess(MyObject *obj, Line *line, int num)
{
	Line   //line1,	// ~̒Sړ~̒S܂ł̐
	       *ld;
	Vector tv,	// ẍꎞxNg
	       va,	// ړ̈ʒuxNg
	       vel,	// cړxNg
	       tmv1,
	       tmv2;
	float tmpf,
	      len1,
	      len2,
	      tm;
	int   i, type, numd;
	
	vel = obj->velocity;

	ld = line;
	numd = num;
	CollisionAreaDelete(obj, &line, &num);	// ̈
	if(num == 0){
		VectorAdd(&obj->xy, &vel);
		return;
	}
		
	
	// 瓖蔻
	vel = obj->velocity;
	while(vel.x!=0 && vel.y!=0){
			obj->xy.y = 0;
		if(vel.x==0 && vel.y==0)		// ړxNgOȂ瓖Ȃ
			break;
		len2 =  1/ VectorLength(&vel);		// ړxNg̒
		tv = vel;
		for(i=0; i<num; i++){
			tm = LineCircleTime(obj, &line[i], &vel);
			if(tm < 0 )	continue;
			VectorMult(&tv, tm);		// Փ˂܂ł̈ړxNg
			VectorAdd(&obj->xy, &tv);	// ՓˉӏɈړ

			// ړxNgՓˌ̌`ɕς
			// QlFhttp://www5f.biglobe.ne.jp/~kenmo/program/buturi5/buturi5.htmlD
			VectorMult(&vel, 1.0f-tm);
			tmv1 = vel;
			VectorRotate90(&tmv1, 1);	// @xNg߂i͋CɂȂj
			VectorNormalize(&tmv1);	// PʃxNg
			VectorInvert(&vel);		// ړxNg]
			VectorMult(&tmv1, VectorDotProduct(&vel, &tmv1));	// @xNĝ΂
			tmv2 = VectorSubNew(&tmv1, &vel);	// sړxNg
			VectorMult(&tmv2, 2.0f);
			VectorAdd(&vel, &tmv2);
			break;
		}
	}
		
}


#endif
